import tempfile
import pandas as pd
from pathlib import Path
from psycopg2 import connect
import psycopg2
from xplordb.datamodel.person import Person
from xplordb.datamodel.company import Company
from xplordb.datamodel.dataset import Dataset
from xplordb.import_data import ImportData
from xplordb.import_ddl import ImportDDL
from xplordb.sqlalchemy.base import create_session


path = Path(tempfile.mkdtemp())


collar_columns = ["data_set", "hole_id", "x", "y", "z", "eoh", "planned_x", "planned_y", "planned_z", "planned_eoh", "dip", "azimuth", "loaded_by", "load_date", "srid", "project_srid"]
survey_columns = ["data_set", "hole_id", "depth_m", "dip", "azimuth_grid", "loaded_by"]
categories_columns = ["table_name", "name", "name_column", "schema"]
person_columns = ["code", "person", "load_date"]
dataset_columns = ["data_set", "full_name", "loaded_by", "load_date"]
profile_columns = ["id", "name", "type", "assay_name", "parameters"]
auto_columns = ["id", "name", "profile", "source_folder", "person", "dataset", "frequency", "log_folder", "creation_date", "last_run_date", "run_count", "failed_run_count"]
assay_column_columns = ["name", "assay", "series_type", "unit", "category_name", "upper_box_column", "lower_box_column", "upper_whisker_column", "lower_whisker_column", 
               "image_format_col", "detection_min_col", "detection_max_col", "display_name", "symbology"]
assay_column_definition_columns = ["name", "variable", "col"]
assay_definition_columns = ["variable", "table_name", "hole_id_col", "x_col", "dataset_col", "person_col", "import_date_col", "x_end_col", "schema"]
assays_columns = ["variable", "domain", "data_extent", "display_name"]


def migrate(conn) -> None:

    # 1st step : export data as csv
    # collars
    with conn.cursor() as cur:
        cur.execute(f"select {','.join(collar_columns)} from dh.collar")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= collar_columns)
        df.to_csv(path / "collars.csv", index=False)

    # metadata
    with conn.cursor() as cur:
        cur.execute("""
                    SELECT column_name
                    FROM information_schema.columns
                    WHERE table_schema = 'dh'
                    AND table_name = 'metadata';
                    """)
        columns = cur.fetchall()
        columns = [col[0] for col in columns]

        cur.execute(f"select {','.join(columns)} from dh.metadata")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= columns)
        df.to_csv(path / "metadata.csv", index=False)

    # metadata types
    with conn.cursor() as cur:
        cur.execute("""
                    SELECT column_name, data_type
                    FROM information_schema.columns
                    WHERE table_schema = 'dh'
                    AND table_name = 'metadata';
                    """)
        types = cur.fetchall()
        df = pd.DataFrame(types, columns= ["column", "type"])
        df.to_csv(path / "metadata_type.csv", index=False)

    # surveys
    with conn.cursor() as cur:
        cur.execute(f"select {','.join(survey_columns)} from dh.surv")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= survey_columns)
        df.to_csv(path / "surveys.csv", index=False)

    # categories
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(categories_columns)} from ref.categories")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= categories_columns)
        df.to_csv(path / "categories.csv", index=False)

    # person
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(person_columns)} from ref.person")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= person_columns)
        df.to_csv(path / "person.csv", index=False)

    # dataset
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(dataset_columns)} from ref.data_sets")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= dataset_columns)
        df.to_csv(path / "datasets.csv", index=False)

    # import profile
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(profile_columns)} from ref.import_profiles")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= profile_columns)
        df.to_csv(path / "import_profiles.csv", index=False)

    # automated imports
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(auto_columns)} from ref.automated_imports")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= auto_columns)
        df.to_csv(path / "automated_imports.csv", index=False)

    # assay metadatas
    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(assay_column_columns)} from assay.assay_column")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= assay_column_columns)
        df.to_csv(path / "assay_column.csv", index=False)

    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(assay_column_definition_columns)} from assay.assay_column_definition")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= assay_column_definition_columns)
        df.to_csv(path / "assay_column_definition.csv", index=False)

    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(assay_definition_columns)} from assay.assay_definition")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= assay_definition_columns)
        df.to_csv(path / "assay_definition.csv", index=False)

    with conn.cursor() as cur:
        
        cur.execute(f"select {','.join(assays_columns)} from assay.assays")
        res = cur.fetchall()
        df = pd.DataFrame(res, columns= assays_columns)
        df.to_csv(path / "assays.csv", index=False)


    # assay raw tables
    assays = pd.read_csv(path / "assay_definition.csv")
    for assay in assays["table_name"]:
        with conn.cursor() as cur:
            cur.execute(f"""
                        SELECT column_name
                        FROM information_schema.columns
                        WHERE table_schema = 'assay'
                        AND table_name = '{assay}';
                        """)
            columns = cur.fetchall()
            columns = [col[0] for col in columns]
            # put column name in double quotes
            quoted_columns = [f'"{column}"' for column in columns]

            cur.execute(f"select {','.join(quoted_columns)} from assay.{assay}")
            res = cur.fetchall()
            df = pd.DataFrame(res, columns= columns)
            filename = "assay_table_" + assay + ".csv"
            df.to_csv(path / filename, index=False)

    # category tables
    categories = pd.read_csv(path / "categories.csv")
    for category in categories["table_name"]:
        with conn.cursor() as cur:
            columns = ["name", "description"]
            cur.execute(f"select {','.join(columns)} from public.{category}")
            res = cur.fetchall()
            df = pd.DataFrame(res, columns= columns)
            filename = "category_table_" + category + ".csv"
            df.to_csv(path / filename, index=False)


    # 2nd step : delete current schema (except public)
    q = """
        SELECT schema_name
        FROM information_schema.schemata
        WHERE schema_name NOT LIKE 'pg_%'
        AND schema_name != 'information_schema';
        """

    with conn.cursor() as cur:
        cur.execute(q)
        schemas = cur.fetchall()
        schemas = [schema[0] for schema in schemas]

    for schema in schemas:
        if schema == "public":
            continue
        with conn.cursor() as cur:
            cur.execute(f"DROP SCHEMA {schema} CASCADE;")
        conn.commit()


    # 3rd step : install new schema
    new_database = "xplordb_migrated"

    import_ddl = ImportDDL(db=new_database,
                        user='postgres',
                        password='larmiche',
                        host='localhost',
                        port=5432)
    import_ddl.import_xplordb_schema()

    # 4th step : import data
    # ref : company, persons, datasets, import_profile, automated_import, categories

    session = create_session(
        database=new_database,
        host='localhost',
        port=5432,
        user='postgres',
        password='larmiche')

    company = Company("openlog")
    import_data = ImportData(session)
    import_data.import_companies_array([company])

    person_df = pd.read_csv(path / "person.csv")
    dataset_df = pd.read_csv(path / "datasets.csv")

    persons = [Person(name=person_name, company=company) for person_name in person_df["person"]]
    trigrams = {row["code"]:row["person"] for _, row in person_df.iterrows()}

    datasets = []
    for _, row in dataset_df.iterrows():
        person_trigram = row["loaded_by"]
        ds_name = row["full_name"]
        person = Person(name=trigrams.get(person_trigram), company=company)
        dataset = Dataset(name=ds_name, person=person)
        datasets.append(dataset)

    import_data.import_persons_array(persons)
    import_data.import_datasets_array(datasets)
    session.commit()

    # new person, dataset IDs
    person_df = pd.read_csv(path / "person.csv")
    person_dict = {}
    for i, person in person_df.iterrows():
        person_dict[person["code"]] = i + 1

    dataset_df = pd.read_csv(path / "datasets.csv")
    dataset_dict = {}
    for i, dataset in dataset_df.iterrows():
        dataset_dict[dataset["data_set"]] = i + 1

    # import profiles
    profiles = pd.read_csv(path / "import_profiles.csv")
    with conn.cursor() as cur:
        for _, profile in profiles.iterrows():
            quoted_columns = [f'"{column}"' for column in profile_columns]
            col_str = ",".join(quoted_columns)
            cur.execute(f"INSERT INTO ref.import_profiles({col_str}) VALUES({','.join(['%s' for col in profile_columns])})",
                        profile.astype(str).to_list())

    conn.commit()

    # automated imports
    autos = pd.read_csv(path / "automated_imports.csv")
    with conn.cursor() as cur:
        for _, auto in autos.iterrows():
            # change person and dataset by IDs
            auto["person"] = person_dict.get(auto["person"])
            auto["dataset"] = dataset_dict.get(auto["dataset"])
            row = auto.to_list()
            row = [str(elt) if not pd.isna(elt) else None for elt in row]

            quoted_columns = [f'"{column}"' for column in auto_columns]
            col_str = ",".join(quoted_columns)
            cur.execute(f"INSERT INTO ref.automated_imports({col_str}) VALUES({','.join(['%s' for col in auto_columns])})",
                        row)
    conn.commit()

    # collars
    collar_df = pd.read_csv(path / "collars.csv")
    with conn.cursor() as cur:
        for _, collar in collar_df.iterrows():
            # change person and dataset by IDs
            collar["loaded_by"] = person_dict.get(collar["loaded_by"])
            collar["data_set"] = dataset_dict.get(collar["data_set"])
            quoted_columns = [f'"{column}"' for column in collar_columns]
            col_str = ",".join(quoted_columns)
            cur.execute(f"INSERT INTO dh.collar({col_str}) VALUES({','.join(['%s' for col in collar_columns])})",
                        collar.astype(str).to_list())
    conn.commit()

    # surveys
    survey_df = pd.read_csv(path / "surveys.csv")
    with conn.cursor() as cur:
        for _, survey in survey_df.iterrows():
            # change person and dataset by IDs
            survey["loaded_by"] = person_dict.get(survey["loaded_by"])
            survey["data_set"] = dataset_dict.get(survey["data_set"])
            quoted_columns = [f'"{column}"' for column in survey_columns]
            col_str = ",".join(quoted_columns)
            cur.execute(f"INSERT INTO dh.surv({col_str}) VALUES({','.join(['%s' for col in survey_columns])})",
                        survey.astype(str).to_list())
    conn.commit()

    # metadata (add columns)
    metadata_df = pd.read_csv(path / "metadata.csv")
    metadata_types = pd.read_csv(path / "metadata_type.csv")
    new_columns = [column for column in metadata_df.columns if column not in  ["hole_id"]]
    with conn.cursor() as cur:
        for new_column in new_columns:
            type_ = metadata_types.loc[metadata_types["column"] == new_column,"type"].values[0]
            cur.execute(f"ALTER TABLE dh.metadata ADD COLUMN {new_column} {type_}")

    conn.commit()

    # metadata (update)
    with conn.cursor() as cur:
        for new_column in new_columns:
            for _, metadata in metadata_df.iterrows():
                hole_id = metadata["hole_id"]
                value = metadata[new_column]
                if pd.isna(value):
                    continue
                cur.execute(f'UPDATE dh.metadata SET "{new_column}" = %s WHERE hole_id = %s',
                            (value, hole_id))

    conn.commit()

    # categories (metadata)
    categories_df = pd.read_csv(path / "categories.csv")
    categories_df = categories_df.replace(pd.NA, None)
    with conn.cursor() as cur:
        for _, category in categories_df.iterrows():
            row = category.to_list()
            row = [elt if not pd.isna(elt) else "" for elt in row]
            cur.execute("INSERT INTO ref.categories VALUES (%s, %s, %s, %s)",
                        row)

    conn.commit()

    # assay (metadatas)
    assays_df = pd.read_csv(path / "assays.csv")
    with conn.cursor() as cur:
        for _, assay in assays_df.iterrows():
            row = assay.to_list()
            cur.execute("INSERT INTO assay.assays VALUES (%s, %s, %s, %s)",
                        row)
    conn.commit()

    assay_def_df = pd.read_csv(path / "assay_definition.csv")
    with conn.cursor() as cur:
        for _, assay in assay_def_df.iterrows():
            row = assay.to_list()
            cur.execute("INSERT INTO assay.assay_definition VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)",
                        row)
    conn.commit()

    assay_col_df = pd.read_csv(path / "assay_column.csv")
    with conn.cursor() as cur:
        for _, assay in assay_col_df.iterrows():
            row = assay.to_list()
            row = [elt if not pd.isna(elt) else "" for elt in row]
            cur.execute("INSERT INTO assay.assay_column VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)",
                        row)
    conn.commit()

    assay_col_def_df = pd.read_csv(path / "assay_column_definition.csv")
    with conn.cursor() as cur:
        for _, assay in assay_col_def_df.iterrows():
            row = assay.to_list()
            cur.execute("INSERT INTO assay.assay_column_definition VALUES (%s, %s, %s)",
                        row)
    conn.commit()

    # category tables (no need to create them because public schema is preserved)

    # assay tables creation

    def sql_assay_table_definition(table_name: str):
        assay = assays_df.loc[assays_df["variable"] == table_name].iloc[0]
        domain = assay["domain"]
        extent = assay["data_extent"]

        # columns description
        columns = assay_col_df.loc[assay_col_df["assay"] == table_name]

        # raw table
        df = pd.read_csv(path / str("assay_table_" + table_name + ".csv"))
        col_sql = []
        for col in df.columns:
            
            if col == "hole":
                type_ = "character varying NOT NULL"
            elif col == "dataset":
                type_ = "integer NOT NULL"
            elif col == "person":
                type_ = "integer REFERENCES ref.person(id) ON DELETE CASCADE"
            elif col == "import_date":
                type_ = "DATE"
            elif col == "x":
                if domain == "depth":
                    type_ = "FLOAT NOT NULL"
                else:
                    type_ = "timestamp without time zone NOT NULL"

            elif col == "x_end":
                if domain == "depth":
                    type_ = "FLOAT NOT NULL"
                else:
                    type_ = "timestamp without time zone NOT NULL"
            
            # measure column
            elif col in columns["name"].to_list():
                column = columns.loc[columns["name"] == col].iloc[0]
                serie_type = column["series_type"]

                if serie_type == "numerical":
                    type_ = "REAL"
                elif serie_type == "polar":
                    type_ = "assay.azimuth"
                elif serie_type in ["nominal", "categorical"]:
                    type_ = "TEXT"
                elif serie_type == "datetime":
                    type_ = "DATE"
                elif serie_type == "spherical":
                    type_ = "assay.spherical_data"
                elif serie_type == "imagery":
                    type_ = "BLOB"
            
            # detection limits or uncertainties
            else:
                type_ = "REAL"
            
            col_sql.append(f'"{col}"' + " " + type_)

        # primary key
        pkey_sql = "PRIMARY KEY (hole, dataset, x)" if extent == "discrete" else "PRIMARY KEY (hole, dataset, x, x_end)"

        # foreign key
        key_sql = "FOREIGN KEY (hole, dataset) REFERENCES dh.collar(hole_id, data_set) ON DELETE CASCADE"

        col_sql.append(pkey_sql)
        col_sql.append(key_sql)
        sql = f'CREATE TABLE assay."{table_name}" ({",".join(col_sql)});'
        return sql

    # get (hole_id, data_set) pairs from dh.collar
    with conn.cursor() as cur:
        cur.execute("select hole_id, data_set from dh.collar")
        collar_pkeys = cur.fetchall()

    # insert data
    with conn.cursor() as cur:
        for _, assay in assays_df.iterrows():
        
            table_name = assay["variable"]
            # create table
            cur.execute(sql_assay_table_definition(table_name))

            # grant permissions to xdb roles
            cur.execute(f'ALTER TABLE assay."{table_name}" OWNER TO xdb_admin;')
            cur.execute(f'GRANT SELECT, TRIGGER ON TABLE assay."{table_name}" TO xdb_viewer;')
            cur.execute(f'GRANT SELECT, UPDATE, INSERT, TRIGGER ON TABLE assay."{table_name}" TO xdb_logger;')
            cur.execute(f'GRANT ALL ON TABLE assay."{table_name}" TO xdb_admin;')


            print("Importing assay table : ", table_name)
            # insert
            df = pd.read_csv(path / str("assay_table_" + table_name + ".csv"))
            placeholders = ",".join(["%s" for i in range(df.shape[1])])
            for _, row in df.iterrows():
                # replace dataset and person by IDs
                row["person"] = person_dict.get(row["person"])
                row["dataset"] = dataset_dict.get(row["dataset"])
                hole_dataset = (row["hole"], row["dataset"])
                row = [elt if not pd.isna(elt) else None for elt in row]
                # some assay have hole_id not in dh.collar !
                if hole_dataset in collar_pkeys:
                    cur.execute(f'INSERT INTO assay."{table_name}" VALUES ({placeholders})',
                            row)
    conn.commit()


import pytest
from psycopg2 import connect


@pytest.fixture()
def new_db():
    return 'test_xplordb'


def test_lite_database_creation(db_creator, new_db, host, port, user, password):
    """
    Test lite database creation

    :param db_creator: fixture to create a database and then remove it from server
    """
    db_creator.create_db(new_db)
    expected_schema_list = ['assay', 'dh', 'display', 'information_schema', 'pg_catalog', 'pg_toast', 'public', 'ref']
    check_schema_list(expected_schema_list, host, new_db, password, port, user)


def check_schema_list(expected_schema_list, host, new_db, password, port, user):
    with connect(database=new_db,  host=host,  port=port,  user=user, password=password) as session:
        with session.cursor() as cur:
            cur.execute("SELECT nspname FROM pg_catalog.pg_namespace ORDER BY nspname")
            schema_list = [res[0] for res in cur.fetchall()]
    assert expected_schema_list == schema_list



from sqlalchemy import Column, String, ForeignKey, Integer
from sqlalchemy.orm import relationship

from xplordb.datamodel.dataset import Dataset
from xplordb.sqlalchemy.base import Base


class XplordbDatasetTable(Dataset, Base):
    """
    Define spatialite columns for Dataset definition
    """
    __tablename__ = "data_sets"
    __table_args__ = {"schema": 'ref'}

    id = Column("id", Integer, primary_key = True)
    name = Column('name', String)
    person = Column('person', Integer, ForeignKey('ref.person.id'))

    collars = relationship("XplordbCollarTable")
    surveys = relationship("XplordbSurveyTable")
    
    person_rel = relationship("XplordbPersonTable", back_populates="datasets")


    def to_dataset(self) -> Dataset:

        person = self.person_rel
        return Dataset(name = self.name, person=person.to_person())


from sqlalchemy import Column, String, Boolean, Integer, ForeignKey
from sqlalchemy.orm import relationship

from xplordb.datamodel.person import Person
from xplordb.datamodel.company import Company

from xplordb.sqlalchemy.base import Base


class XplordbPersonTable(Person, Base):
    """
    Define Xplordb columns for Person definition
    """
    __tablename__ = "person"
    __table_args__ = {"schema": 'ref'}

    id = Column('id', Integer, primary_key=True)
    name = Column('name', String)
    company = Column("company", String, ForeignKey('ref.company.name'))
    
    company_rel = relationship("XplordbCompanyTable", back_populates="persons")
    datasets = relationship("XplordbDatasetTable", back_populates="person_rel")

    def to_person(self) -> Person:

        return Person(name=self.name,
                      company=Company(name=self.company))


"""
Minimal data acces object for xplordb test data injection

Minimal data acces object for xplordb test data injection.
With a pyscog2 session, insertion of :
- person
- dataset
- hole type
- collar
- survey



xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/02"
__license__ = "AGPLv3"

from sqlalchemy import text

from xplordb.datamodel.collar import Collar
from xplordb.datamodel.dataset import Dataset
from xplordb.datamodel.person import Person
from xplordb.datamodel.company import Company
from xplordb.datamodel.survey import Survey
from xplordb.import_data import ImportData


class MinimalDao:

    def __init__(self, db_session):
        """
        Simple Data Access Object class to insert some basic data into xplordb database
        :param db_session: psycog2 connection
        """
        self._db_session = db_session
        self._importer = ImportData(db_session)

    def create_company(self, name: str):
        self._importer.import_companies_array([Company(name)])

    def create_person(self, name: str, company: str):
        self._importer.import_persons_array([Person(name, Company(company))])

    def create_dataset(self, name: str, person: str, company: str):
        self._importer.import_datasets_array([Dataset(name, Person(person, Company(company)))])

    def create_collar(self, name: str, data_set: str, person: str, company: str,
                      x: float = 0.0, y: float = 0.0, z: float = 0.0, srid: int = 3857, eoh: float = None,
                      create_needed_data: bool = True):
        if create_needed_data:
            self.create_company(company)
            self.create_person(person, company)
            self.create_dataset(data_set, person, company)
           
        person_obj = Person(person, Company(company))
        dataset_obj = Dataset(data_set, Person(person, Company(company)))
        self._importer.import_collars_array([Collar(name, dataset_obj, person_obj, x, y, z, srid, eoh)])

    def create_surv(self, depth: float, dip: float, azimuth: float, colar: str, data_set: str, person: str, company: str):
        person_obj = Person(person, Company(company))
        dataset_obj = Dataset(data_set, Person(person, Company(company)))
        self._importer.import_surveys_array([Survey(colar, dataset_obj, person_obj, depth, dip, azimuth)])



from sqlalchemy import Column, String, ForeignKey, Integer
from sqlalchemy.orm import relationship

from xplordb.datamodel.company import Company
from xplordb.sqlalchemy.base import Base


class XplordbCompanyTable(Company, Base):
    """
    Define spatialite columns for Company definition
    """
    __tablename__ = "company"
    __table_args__ = {"schema": 'ref'}

    name = Column("name", String, primary_key = True)

    persons = relationship("XplordbPersonTable", back_populates="company_rel")


    def to_company(self) -> Company:

        return Company(name=self.name)


from xplordb.datamodel.person import Person

"""
Module define a data set

Module define a data set


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/09"
__license__ = "AGPLv3"


class Dataset:

    def __init__(self, name: str, person: Person):
        """
        Define a data set used to store collar and associated data

        :param name: data set unique identifier
        :param person: person used to load data set
        """
        self.name = name
        self.person = person



"""
Module define a collar

Module define a collar


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["jmkerloch"]
__contact__ = "geology@oslandia.com"
__date__ = "2022/02/09"
__license__ = "AGPLv3"

from datetime import datetime
from xplordb.datamodel.dataset import Dataset
from xplordb.datamodel.person import Person


class Collar:

    def __init__(self, hole_id: str, data_set: Dataset, loaded_by: Person,
                 x: float, y: float, z: float, srid: int, 
                 eoh: float = None, planned_x: float = None, planned_y: float = None, planned_z: float = None,planned_eoh: float = None, 
                 project_srid: int = 4326, dip: float = -90, azimuth: float = 0
                 ):
        """
        Define a collar

        :param hole_id: collar name and unique identifier
        :param data_set: data set associated to collar
        :param loaded_by: person used to load collar
        :param x: x coordinate of collar
        :param y: y coordinate of collar
        :param z: z coordinate of collar
        :param srid: srid for coordinate
        :param eoh : (optionnal) end of hole depth
        """
        self.hole_id = hole_id
        self.data_set = data_set
        self.loaded_by = loaded_by
        self.x = x
        self.y = y
        self.z = z
        self.srid = srid
        self.project_srid = project_srid
        self.eoh = eoh
        self.planned_x = planned_x
        self.planned_y = planned_y
        self.planned_z = planned_z
        self.planned_eoh = planned_eoh
        self.dip = dip
        self.azimuth = azimuth



from xplordb.migration import v1
from packaging.version import Version

# key : current version to be upgraded
MIGRATIONS = {"0.13.8" : v1.migrate}

def migrate_database(conn, target_version: str):
    """
    Migrate iteratively.  
    User must be connected as administrator or database owner.
    Each migration is destructive, with the following steps:  
    - export raw data as csv
    - remove old schema
    - install new schema
    - import data
    """
    cur = conn.cursor()

    # get current version
    cur.execute("SELECT version FROM ref.xplordb")
    version_str = cur.fetchone()[0]
    version = Version(version_str)
    if version < Version("0.13.8"):
        print("Xplordb schema is too old. Aborting")
        return
    target_version = Version(target_version)

    while version < target_version:
        migrate = MIGRATIONS[version_str]
        migrate(conn)
        conn.commit()
        # get new version
        cur.execute("SELECT version FROM ref.xplordb")
        version_str = cur.fetchone()[0]
        version = Version(version_str)


from sqlalchemy import Column, String, ForeignKey, Float, Numeric, ForeignKeyConstraint, Integer

from xplordb.datamodel.survey import Survey

from xplordb.sqlalchemy.base import Base


class XplordbSurveyTable(Survey, Base):
    """
    Define Xplordb columns for Survey definition
    """
    __tablename__ = "surv"

    data_set = Column("data_set", Integer, ForeignKey('ref.data_sets.id'), primary_key=True)
    hole_id = Column("hole_id", String, primary_key=True)
    loaded_by = Column("loaded_by", Integer,  ForeignKey('ref.person.id'))

    # Column defined as Numeric so they can be used as primary keys
    # Value in Survey class defined as float : not used as decimal
    depth = Column("depth_m", Numeric(asdecimal=False), primary_key=True)
    dip = Column("dip", Float)
    azimuth = Column("azimuth_grid", Float)

    __table_args__ = (
        ForeignKeyConstraint(
            ["hole_id", "data_set"],
            ["dh.collar.hole_id", "dh.collar.data_set"],
            ondelete="CASCADE"
        ),
        {"schema": "dh"}
    )


"""
Module define a collar

Module define a collar


xplordb

Copyright (C) 2022  Oslandia / OpenLog
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__authors__ = ["vlarmet"]
__contact__ = "vincent.larmet@apeiron.technology"
__date__ = "2024/04/11"
__license__ = "AGPLv3"

from datetime import datetime
from xplordb.datamodel.dataset import Dataset


class RawCollarMetadata:

    def __init__(self, hole_id: str, data_set: Dataset, extra_cols: dict
                 ):
        """
        Define a raw collar metadata.  
        Metadata are additional collar attributes.

        :param hole_id: collar name
        :param data_set: collar's dataset
        :param extra_cols
        """
        self.hole_id = hole_id
        self.data_set = data_set
        self.extra_cols = extra_cols
    
class CollarMetadata:

    def __init__(self, hole_id: str, data_set: Dataset
                 ):
        """
        Define a collar metadata.  
        Metadata are additional collar attributes.

        :param hole_id: collar name
        :param data_set: collar's dataset
        :param extra_cols
        """
        self.hole_id = hole_id
        self.data_set = data_set

def create_collar_metadata(raw_collar_metadata: RawCollarMetadata) -> CollarMetadata:

    hole_id = raw_collar_metadata.hole_id
    data_set = raw_collar_metadata.data_set
    metadata = CollarMetadata(hole_id=hole_id, data_set=data_set)
    for col_name, value in raw_collar_metadata.extra_cols.items():
        setattr(metadata, col_name, value)

    return metadata


from datetime import datetime

import pytest
from sqlalchemy import text

from xplordb.datamodel.collar import Collar
from xplordb.datamodel.metadata import RawCollarMetadata
from xplordb.datamodel.company import Company
from xplordb.datamodel.dataset import Dataset
from xplordb.datamodel.person import Person
from xplordb.datamodel.survey import Survey
from xplordb.import_data import ImportData
from tests.conftest import COMPANY, PERSON, DATASET, COMPANY_B, COMPANY_B_PERSON, COMPANY_B_DATASET



def test_metadata_import(data_importer_with_dataset, db_session):

    # get dataset ID
    dataset_id = data_importer_with_dataset._get_dataset_id(DATASET)
    # add some collars
    collars = [Collar('collar', DATASET, PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,1000),
               Collar('collar2', DATASET, PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,None),
               Collar('collar3', DATASET, PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,None, None)]
    data_importer_with_dataset.import_collars_array(collars)
    data_importer_with_dataset.commit()

    # import metadata
    metadatas = [RawCollarMetadata(hole_id="collar", data_set=DATASET, extra_cols={"col_a" : 10}),
                 RawCollarMetadata(hole_id="collar3", data_set=DATASET, extra_cols={"col_b" : 5})]
    
    data_importer_with_dataset.import_metadatas_array(metadatas)
    data_importer_with_dataset.commit()
    expected = [(dataset_id, "collar", 10, None), 
                (dataset_id, "collar2", None, None), 
                (dataset_id, "collar3", None, 5)]
    
    with db_session.cursor() as cur:
        cur.execute("SELECT * FROM dh.metadata ORDER BY hole_id")
        values = cur.fetchall()
        assert len(values) == len(collars)
        for value, exp in zip(values, expected):
            assert value == exp

def test_duplicated_company_import(data_importer, db_session):
    """
    Test duplicated company import
    """
    companies = [Company("company_a"), Company("company_b"), Company("company_b")]

    with pytest.raises(ImportData.ImportException):
        data_importer.import_companies_array(companies)
        data_importer.commit()


def test_duplicated_person_import(data_importer, db_session):
    """
    Test duplicated person import
    """
    company = Company("company_a")
    data_importer.import_companies_array([company])

    persons = [Person('a', company),
               Person('b', company),
               Person('b', company) ]
    with pytest.raises(ImportData.ImportException):
        data_importer.import_persons_array(persons)
        data_importer.commit()


def test_person_import(data_importer, db_session):
    """
    Test person import
    """
    company = Company("company_a")
    data_importer.import_companies_array([company])

    persons = [Person('a', company),
               Person('b', company),
               Person('c', company) ]
    data_importer.import_persons_array(persons)
    data_importer.commit()

    with db_session.cursor() as cur:
        cur.execute("SELECT id, name, company FROM ref.person")
        values = cur.fetchall()
        assert len(values) == len(persons)
        for i in range(0, len(values)):
            assert values[i][0] == i + 1
            assert values[i][1] == persons[i].name
            assert values[i][2] == persons[i].company.name


def test_get_available_persons(data_importer, db_session):
    """
    Test person read from database
    """
    test_person_import(data_importer, db_session)
    persons = data_importer.get_available_persons()

    assert [person.name for person in persons] == ["a", "b", "c"]
    assert [person.company.name for person in persons] == ["company_a", "company_a", "company_a"]


def test_dataset_import_invalid_person(data_importer_with_dataset):
    """
    Test dataset import with invalid person
    """

    with pytest.raises(ImportData.ImportException):
        datasets = [Dataset('test', Person("invalid", COMPANY))]
        data_importer_with_dataset.import_datasets_array(datasets)
        data_importer_with_dataset.commit()


def test_dataset_import(data_importer, db_session):
    """
    Test dataset import
    """
    company = Company("company_a")
    data_importer.import_companies_array([company])

    person = Person('xdb', company)
    data_importer.import_persons_array([person])

    datasets = [Dataset('test', person),
                Dataset('test2', person), ]
    
    data_importer.import_datasets_array(datasets)
    data_importer.commit()

    person_id = data_importer._get_person_id(PERSON)
    with db_session.cursor() as cur:
        cur.execute("SELECT id, name, person FROM ref.data_sets")
        values = cur.fetchall()
        assert len(values) == len(datasets)
        for i in range(0, len(values)):
            assert values[i][0] == i + 1
            assert values[i][1] == datasets[i].name
            assert values[i][2] == person_id

def test_duplicated_dataset_import(data_importer, db_session):
    """
    Test dataset import
    """
    company = Company("company_a")
    data_importer.import_companies_array([company])

    person = Person('xdb', company)
    data_importer.import_persons_array([person])

    datasets = [Dataset('test', person),
                Dataset('test', person), ]

    with pytest.raises(ImportData.ImportException):
        data_importer.import_datasets_array(datasets)
        data_importer.commit()



def test_get_available_datasets(data_importer, db_session):
    """
    Test datasets read from database
    """
    test_dataset_import(data_importer, db_session)
    datasets = data_importer.get_available_datasets()
    assert [dataset.name for dataset in datasets] == ['test', 'test2']
    assert [dataset.person.name for dataset in datasets] == ['xdb', 'xdb']
    assert [dataset.person.company.name for dataset in datasets] == ['company_a', 'company_a']


def test_collar_import_invalid_fk(data_importer_with_dataset):
    """
    Test collar import with an invalid foreign key
    """

    with pytest.raises(ImportData.ImportException):
        collar = [Collar('collar',Dataset('test2', PERSON), PERSON, 100.0, 100.0, 0.0, 3857), ]
        data_importer_with_dataset.import_collars_array(collar)
        data_importer_with_dataset.commit()


def test_collar_import(data_importer_with_dataset, db_session):
    """
    Test collar import
    """

    collars = [Collar('collar', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,1000),
               Collar('collar2', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857),
               Collar('collar3', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, None, None, None,None, None)]
    data_importer_with_dataset.import_collars_array(collars)
    data_importer_with_dataset.commit()

    person_id = data_importer_with_dataset._get_person_id(PERSON)
    dataset_id = data_importer_with_dataset._get_dataset_id(DATASET)

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,hole_id,loaded_by,x,y,z,srid,eoh, planned_x, planned_y, planned_z, planned_eoh FROM dh.collar")
        values = cur.fetchall()
        assert len(values) == len(collars)
        for i in range(0, len(values)):
            assert values[i][0] == dataset_id
            assert values[i][1] == collars[i].hole_id
            assert values[i][2] == person_id
            assert values[i][3] == collars[i].x
            assert values[i][4] == collars[i].y
            assert values[i][5] == collars[i].z
            assert values[i][6] == collars[i].srid
            assert values[i][7] == collars[i].eoh
            assert values[i][8] == collars[i].planned_x
            assert values[i][9] == collars[i].planned_y
            assert values[i][10] == collars[i].planned_z
            assert values[i][11] == collars[i].planned_eoh

def test_duplicated_collar_import(data_importer_with_dataset, db_session):
    """
    Test duplicated collar import
    """

    collars = [Collar('collar', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,1000),
               Collar('collar2', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857),
               Collar('collar2', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, None, None, None,None, None)]
    with pytest.raises(ImportData.ImportException):
        data_importer_with_dataset.import_collars_array(collars)
        data_importer_with_dataset.commit()



def test_survey_import_invalid_fk(data_importer_with_dataset):
    """
    Test survey import with an invalid foreign key
    """
    
    with pytest.raises(ImportData.ImportException):
        surveys = [Survey('collar', Dataset('test2', PERSON), PERSON, 0.0, 45.0, 0.0), ]
        data_importer_with_dataset.import_surveys_array(surveys)
        data_importer_with_dataset.commit()


def test_survey_import(data_importer_with_dataset, db_session):
    """
    Test survey import
    """

    collars = [Collar('collar', DATASET, PERSON, 100.0, 100.0, 0.0, 3857), ]
    data_importer_with_dataset.import_collars_array(collars)

    surveys = [Survey('collar', DATASET, PERSON, 0.0, 45.0, 0.0), ]
    data_importer_with_dataset.import_surveys_array(surveys)
    data_importer_with_dataset.commit()

    person_id = data_importer_with_dataset._get_person_id(PERSON)
    dataset_id = data_importer_with_dataset._get_dataset_id(DATASET)

    with db_session.cursor() as cur:
        cur.execute("SELECT data_set,hole_id,loaded_by,depth_m,dip,azimuth_grid FROM dh.surv")
        values = cur.fetchall()
        assert len(values) == len(surveys)
        for i in range(0, len(values)):
            assert values[i][0] == dataset_id
            assert values[i][1] == surveys[i].hole_id
            assert values[i][2] == person_id
            assert values[i][3] == surveys[i].depth
            assert values[i][4] == surveys[i].dip
            assert values[i][5] == surveys[i].azimuth

def test_duplicated_survey_import(data_importer_with_dataset, db_session):
    """
    Test duplicated survey import
    """

    collars = [Collar('collar', DATASET, PERSON, 100.0, 100.0, 0.0, 3857), ]
    data_importer_with_dataset.import_collars_array(collars)
    surveys = [Survey('collar', DATASET, PERSON, 0.0, 45.0, 0.0), Survey('collar', DATASET, PERSON, 2.0, 45.0, 0.0), Survey('collar', DATASET, PERSON, 2.0, 45.0, 0.0)]
    with pytest.raises(ImportData.ImportException):
        data_importer_with_dataset.import_surveys_array(surveys)
        data_importer_with_dataset.commit()

@pytest.fixture()
def test_multi_company_collar_import(data_importer_with_two_companies, db_session):
    """
    Test collar import with same hole ids.
    """
    collars = [
        # company A
        Collar('collar', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,1000),
        Collar('collar2', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857),
        Collar('collar3', DATASET,  PERSON, 100.0, 100.0, 0.0, 3857, None, None, None,None, None),
        # company B
        Collar('collar', COMPANY_B_DATASET,  COMPANY_B_PERSON, 100.0, 100.0, 0.0, 3857, 100.0, 100.0, 0.0,1000),
        Collar('collar2', COMPANY_B_DATASET,  COMPANY_B_PERSON, 100.0, 100.0, 0.0, 3857),
        Collar('collar3', COMPANY_B_DATASET,  COMPANY_B_PERSON, 100.0, 100.0, 0.0, 3857, None, None, None,None, None)
        ]
    data_importer_with_two_companies.import_collars_array(collars)
    data_importer_with_two_companies.commit()

    yield data_importer_with_two_companies


@pytest.fixture()
def test_multi_company_surv_import(test_multi_company_collar_import, db_session):
    """
    Test collar import with same hole ids.
    """
    surveys = [
    # company A
    Survey('collar', DATASET, PERSON, 0.0, 45.0, 0.0),
    Survey('collar', DATASET, PERSON, 2.0, 45.0, 0.0),
    # company B
    Survey('collar', COMPANY_B_DATASET, COMPANY_B_PERSON, 0.0, 45.0, 0.0),
    Survey('collar', COMPANY_B_DATASET, COMPANY_B_PERSON, 2.0, 45.0, 0.0)
    ]
    test_multi_company_collar_import.import_surveys_array(surveys)
    test_multi_company_collar_import.commit()

    yield test_multi_company_collar_import


def test_multi_company_delete_collar(test_multi_company_surv_import, db_session):
    
    dataset_id = test_multi_company_surv_import._get_dataset_id(DATASET)

    # delete collar from first company, check deleted items from cascade
    with db_session.cursor() as cur:
        cur.execute(f"DELETE FROM dh.collar WHERE data_set = {dataset_id} AND hole_id = 'collar'")
        db_session.commit()
        # 1 row removed
        cur.execute(f"select * from dh.collar")
        res = cur.fetchall()
        assert len(res) == 5

        # 2 rows removed
        cur.execute(f"select * from dh.surv")
        res = cur.fetchall()
        assert len(res) == 2

        # 1 row removed
        cur.execute(f"select * from dh.metadata")
        res = cur.fetchall()
        assert len(res) == 5

        # 1 row removed
        cur.execute(f"select * from display.display_collar")
        res = cur.fetchall()
        assert len(res) == 5


def test_multi_company_delete_dataset(test_multi_company_surv_import, db_session):
    
    dataset_id = test_multi_company_surv_import._get_dataset_id(DATASET)

    # delete collar from first company, check surveys have been deleted (cascade)
    with db_session.cursor() as cur:
        cur.execute(f"DELETE FROM ref.data_sets WHERE id = {dataset_id}")
        db_session.commit()
        # 3 row removed
        cur.execute(f"select * from dh.collar")
        res = cur.fetchall()
        assert len(res) == 3

        # 2 rows removed
        cur.execute(f"select * from dh.surv")
        res = cur.fetchall()
        assert len(res) == 2

        # 3 row removed
        cur.execute(f"select * from dh.metadata")
        res = cur.fetchall()
        assert len(res) == 3

        # 3 row removed
        cur.execute(f"select * from display.display_collar")
        res = cur.fetchall()
        assert len(res) == 3

def test_multi_company_delete_person(test_multi_company_surv_import, db_session):
    
    person_id = test_multi_company_surv_import._get_person_id(PERSON)

    # delete collar from first company, check surveys have been deleted (cascade)
    with db_session.cursor() as cur:
        cur.execute(f"DELETE FROM ref.person WHERE id = {person_id}")
        db_session.commit()
        # 3 row removed
        cur.execute(f"select * from dh.collar")
        res = cur.fetchall()
        assert len(res) == 3

        # 2 rows removed
        cur.execute(f"select * from dh.surv")
        res = cur.fetchall()
        assert len(res) == 2

        # 3 row removed
        cur.execute(f"select * from dh.metadata")
        res = cur.fetchall()
        assert len(res) == 3

        # 3 row removed
        cur.execute(f"select * from display.display_collar")
        res = cur.fetchall()
        assert len(res) == 3

def test_multi_company_delete_company(test_multi_company_surv_import, db_session):

    # delete collar from first company, check surveys have been deleted (cascade)
    with db_session.cursor() as cur:
        cur.execute(f"DELETE FROM ref.company WHERE name = '{COMPANY.name}'")
        db_session.commit()
        # 3 row removed
        cur.execute(f"select * from dh.collar")
        res = cur.fetchall()
        assert len(res) == 3

        # 2 rows removed
        cur.execute(f"select * from dh.surv")
        res = cur.fetchall()
        assert len(res) == 2

        # 3 row removed
        cur.execute(f"select * from dh.metadata")
        res = cur.fetchall()
        assert len(res) == 3

        # 3 row removed
        cur.execute(f"select * from display.display_collar")
        res = cur.fetchall()
        assert len(res) == 3
